// based off of
// async_client.cpp
// ~~~~~~~~~~~~~~~~
//
// Copyright (c) 2003-2012 Christopher M. Kohlhoff (chris at kohlhoff dot com)
//
// Distributed under the Boost Software License, Version 1.0. (See accompanying
// file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
//

#ifndef VERSION_CHECKER_HPP
#define VERSION_CHECKER_HPP

#include <iostream>
#include <sstream>
#include <string>

std::string getVersionMessage();

#endif // VERSION_CHECKER_HPP