#ifndef IO_UTILS
#define IO_UTILS

#include <cstddef>
#include <cstdint>

namespace ioutils {

const char RESET_COLOR[] = "\x1b[0;0m";
const char SET_GREEN[] = "\x1b[0;32m";
const char SET_RED[] = "\x1b[0;31m";

} // namespace ioutils

#endif // IO_UTILS
