#define CATCH_CONFIG_MAIN  // This tells Catch to provide a main() - only do this in one cpp file
#include "catch.hpp"

bool verbose=false; // Apparently, we *need* this (OSX)
//#include "GCSampleTests.cpp"
//#include "LibraryTypeTests.cpp"
//#include "KmerHistTests.cpp"

